// $Id: main.cc,v 1.12 2005/06/13 06:06:53 weesan Exp $

#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include "node.h"
#include "wireless_node.h"
#include "link.h"
#include "simulation.h"
#include "torrent_node.h"
#include "debug.h"

#define WIRELESS_SIMULATION
//#define RANDOM_DEPLOYMENT
//#define RANDOM_SEED

#define DEFAULT_GRID_SIZE   16
#define NODE_SPACING        15     // in feet

#define SIM_DIM_X           300    // in feet
#define SIM_DIM_Y           300    // in feet
#define SIM_TOTAL_TIME      1800   // 30 mins

#define USAGE() {                          \
    fprintf(stderr, pzcUsage, pzcProgram); \
    return (1);                            \
}

static int iGridSize = DEFAULT_GRID_SIZE;
static bool bReqSuppression = true;
static bool bLocalRecovery = false;
static bool bOnDemandCaching = false;
static bool bRawFormat = false;
static bool bPushMore = false;

static const char pzcUsage[] = 
"\n"
"Usage: %s [options]\n"
"where options are as below\n"
"\t-h\tHelp\n"
"\t-d\tClass-based debug\n"
"\t-g\tGrid size (Default: 16)\n"
"\t-s\tTurn off REQ suppression (Default: ON)\n"
"\t-l\tTurn on Local Recovery (Default: OFF)\n"
"\t-o\tTurn on on-demand packet caching (Default: OFF)\n"
"\t-r\tDisplay result in raw format for plotting (Default: OFF)\n"
"\t-p\tPush more DAT if available (Default: OFF)\n"
"\n";

static void node2node(void *pvSrcNode, void *pvDstNode, void *pvPacket) {
    Node *pcSrcNode = (Node *)pvSrcNode;
    Node *pcDstNode = (Node *)pvDstNode;
    Packet *pcPacket = (Packet *)pvPacket;

    pcSrcNode->tx(pcPacket, pcDstNode);
}

static void wire_simulation(void) {
    Simulation cSimulation(SIM_TOTAL_TIME);

    // Allocate 3 nodes
    Node *n0 = new Node;
    Node *n1 = new Node(new QueueDropTail);
    Node *n2 = new Node(new QueueDropTail(QUEUE_SIZE));

    // Allocate 4 directional links
    Link *l0 = new Link(n0, n1, LINK_T1, 10);
    Link *l1 = new Link(n1, n0, LINK_T1, 10);
    Link *l2 = new Link(n1, n2, LINK_T1, 10);
    Link *l3 = new Link(n2, n1, LINK_T1, 10);

    // Add default route
    n0->addDefaultRoute(l0);
    n1->addDefaultRoute(l2);

    // Schedule traffic to be sent
    for (int i = 0; i < 17; i++) {
	AT(i * 10, node2node, n0, n2, new Packet(n0->id(), n2->id(), PT_UDP));
    }

    cSimulation.run();
}

static void wireless_display(void) {
    // Display node type
    if (!bRawFormat) {
	printf("\nImage type distribution:\n");
	int piNodeType[TN_TYPE_MAX];
	for (int i = 0; i < TN_TYPE_MAX; i++) {
	    piNodeType[i] = 0;
	}
	
	for (int i = 0; i < cNodes.size(); i++) {
	    TorrentNode *pcNode = (TorrentNode *)cNodes[i];
	    piNodeType[pcNode->nodeType()]++;
	}
	for (int i = 0; i < TN_TYPE_MAX; i++) {
	    printf("Type %d: %d\n", i, piNodeType[i]);
	}
	
	// Display topology
	printf("\nTopology:\n");
	for (int i = 0; i < cNodes.size(); i++) {
	    TorrentNode *pcNode = (TorrentNode *)cNodes[i];
	    printf("%3d(%d) ", pcNode->id(), pcNode->nodeType());
	    if ((i + 1) % iGridSize == 0) {
		printf("\n");
	    }
	}
    }

    // Display some nodes stats
    unsigned uiTotalCompletionTime = 0;
    int iTotalReq = 0;
    int iTotalIncompleted = 0;
    if (!bRawFormat) {
	printf("\nNodes Stats:\n");
    }
    for (int i = 0; i < cNodes.size(); i++) {
	TorrentNode *pcNode = (TorrentNode *)cNodes[i];

	uiTotalCompletionTime += pcNode->done();
	iTotalReq += pcNode->noofReq();

	if (bRawFormat) {
	    printf("%3d\t%d\t%d\t%d\t%d\n",
		   pcNode->id(),
		   pcNode->nodeType(),
		   pcNode->noofReq(),
		   pcNode->fwdedPkts()->maxCacheSize(),
		   pcNode->done());
	    if (pcNode->id() != 0 && !pcNode->done()) {
		iTotalIncompleted++;
	    }
	} else {
	    printf("Node %3d\n", pcNode->id());
	    printf("\ttype: %d\n", pcNode->nodeType());
	    printf("\t# of Req: %d\n", pcNode->noofReq());
	    printf("\tcache size: %d\n", pcNode->fwdedPkts()->maxCacheSize());
	    printf("\tcompletion time: ");
	    if (pcNode->done()) {
		printf("%d\n", pcNode->done());
	    } else {
		// Exclude the sink
		if (pcNode->id() != 0) {
		    iTotalIncompleted++;
		}
		printf("%d (page %d cell %d)\n",
		       pcNode->done(), pcNode->page(), pcNode->cell());
	    }
	}
    }

    printf("\n");
    printf("Ave. completion time: %.02f\n",
	   (float)uiTotalCompletionTime / cNodes.size());
    printf("Total REQ for all nodes: %d\n", iTotalReq);
    printf("Total node with incompleted image: %d\n", iTotalIncompleted);
    printf("Simulation end time: %d, event size = %d\n",
	   NOW(), Simulation::scheduleSize());
    printf("-----\n");
}

static void wireless_simulation(void) {
    Simulation cSimulation(SIM_DIM_X, SIM_DIM_Y, SIM_TOTAL_TIME);

    for (int i = 0; i < iGridSize; i++) {
	for (int j = 0; j < iGridSize; j++) {
#ifdef RANDOM_DEPLOYMENT
	    new TorrentNode(i * NODE_SPACING, j * NODE_SPACING,
			    random() % TN_TYPE_MAX,
			    bReqSuppression,
			    bLocalRecovery,
			    bPushMore,
			    bOnDemandCaching ? new TorrentCache : NULL):
#else
	    new TorrentNode(i * NODE_SPACING, j * NODE_SPACING,
			    i == 0 || j == 0 || 
			    i == iGridSize - 1 || j == iGridSize - 1 
			    ? 0 : 1,
			    bReqSuppression,
			    bLocalRecovery,
			    bPushMore,
			    bOnDemandCaching ? new TorrentCache : NULL);
#endif
	}
    }

    cSimulation.run();

    wireless_display();
}

static void signal_handler(int iSignal) {
    wireless_display();
    if (iSignal == SIGINT) {
	exit(1);
    }
}

int main(int argc, char *argv[]) {
    const char *pzcProgram = argv[0];
    int iOpt;

#ifdef WIRELESS_SIMULATION
    // Setup signal handle
    signal(SIGINT, signal_handler);
    signal(SIGUSR1, signal_handler);
#endif
    
    // Parse the command line options
    while ((iOpt = getopt(argc, argv, "hd:sg:lorp")) != EOF) {
	switch (iOpt) {
	case 'd':
	    Debug::on(optarg);
	    break;
	case 's':
	    bReqSuppression = false;
	    break;
	case 'g':
	    iGridSize = atoi(optarg);
	    break;
	case 'l':
	    bLocalRecovery = true;
	    break;
	case 'o':
	    bOnDemandCaching = true;
	    break;
	case 'r':
	    bRawFormat = true;
	    break;
	case 'p':
	    bPushMore = true;
	    break;
	case 'h':
	    // Fall thru
	default:
	    USAGE();
	}
    }

    argc -= optind;
    argv += optind;

#ifdef RANDOM_SEED
    srand(time(NULL));
#endif

#ifdef WIRELESS_SIMULATION
    wireless_simulation();
#else
    wire_simulation();
#endif
}
