// $Id: node.cc,v 1.3 2005/06/11 18:06:08 weesan Exp $

#include <stdio.h>
#include <assert.h>
#include "node.h"
#include "link.h"
#include "packet.h"
#include "routing_table.h"
#include "simulation.h"

DEBUG_REGISTER(Node);

// Global variables
Nodes cNodes;

///////////////////////////////////

// Static variables
node_id_t Node::tNextId = 0;

///////////////////////////////////

static void nodeRemoveStaleRouteEntries(void *pvNode, void *, void *) {
    Node *pcNode = (Node *)pvNode;

    // Remove stale route entries
    pcNode->routingTable()->removeStaleEntries(NOW());

    // Schedule itself again
    AT(SIM_1_SEC, nodeRemoveStaleRouteEntries, pcNode); 
}

///////////////////////////////////

Node::Node(Queue *_pcQueue) :
    Object("Node"),
    tId(nextId()),
    pcQueue(_pcQueue),
    pcRoutingTable(NULL) {
    // Put into node collection
    cNodes.push_back(this);

    // Allocate a sink if no queue is specified
    if (pcQueue == NULL) {
	pcQueue = new Queue;
	assert(pcQueue != NULL);
    }

    // Allocate a routing table
    pcRoutingTable = new RoutingTable;
    assert(pcRoutingTable != NULL);

#ifdef ROUTE_MAINTENANCE
    // Route maintenance
    AT(SIM_1_SEC, nodeRemoveStaleRouteEntries, this);
#endif

    DEBUG("%d created", id());
}

Node::~Node(void) {
    if (pcQueue) {
	delete pcQueue;
    }
    DEBUG("%d destroyed", id());
}

void Node::addDefaultRoute(Link *pcDefault) {
    routingTable()->addEntry(RT_DEFAULT_ROUTE, pcDefault, RouteEntry::STATIC);
}

void Node::tx(Packet *pcPacket, Link *pcOutgoing) {
    if (validLink(pcOutgoing)) {
	DEBUG("%d tx packet %d to node %d via link %d",
	      id(), pcPacket->id(),
	      pcOutgoing->dstNode()->id(), pcOutgoing->id());
	pcOutgoing->rx(pcPacket, this);
    } else {
	DEBUG("%d failed to find link %d", id(), pcOutgoing->id());
	pcPacket->destroy();
    }
}

void Node::tx(Packet *pcPacket, Node *pcDstNode, Link *pcIncoming) {
    // Lookup the routing table based on dst node id
    Link *pcOutgoing = routingTable()->lookup(pcDstNode->id());
    // If not found, try default route
    if (pcOutgoing == NULL) {
	pcOutgoing = routingTable()->lookup(RT_DEFAULT_ROUTE);
    }
    if (pcOutgoing) {
	// If pcIncoming is NULL, the packet is originated from the node
	if (pcIncoming && *pcOutgoing == *pcIncoming) {
	    DEBUG("%d packet %d comes in the wrong interface %d",
		  pcPacket->id(), pcOutgoing->id());
	    pcPacket->destroy();
	} else {
	    tx(pcPacket, pcOutgoing);
	}
    } else {
	DEBUG("%d failed to find route to next hop %d", id(), pcDstNode->id());
	pcPacket->destroy();
    }
}

void Node::rx(Packet *pcPacket, Link *pcIncoming) {
    DEBUG("%d rx packet %d from node %d via link %d",
	  id(), pcPacket->id(), pcIncoming->srcNode()->id(), pcIncoming->id());

    // If the packet is not for me, try to forward it
    if (!pcPacket->forMe(this)) {
	tx(pcPacket, cNodes[pcPacket->dst()], pcIncoming);
	return;
    }

    // The packet is for me
    if (queue()) {
	queue()->add(pcPacket);
    } else {
	DEBUG("%d queue erorr, drop packet %d", id(), pcPacket->id());
	pcPacket->destroy();
    }
}

///////////////////////////////////

#ifdef DEBUG_NODE

int main(void) {
    for (int i = 0; i < 1000; i++) {
	new Node;
    }
    for (int i = 0; i < cNodes.size(); i++) {
	printf("Node %d, id = %d\n", i, cNodes[i]->id());
    }
}

#endif // DEBUG_NODE
