// $Id: packet.h,v 1.7 2005/06/12 09:31:10 weesan Exp $

#ifndef PACKET_H
#define PACKET_H

#include <list>
#include "node.h"
#include "debug.h"

#define PT_IP                  0
#define PT_ICMP                1
#define PT_IGMP                2
#define PT_IPIP                4  
#define PT_TCP                 6
#define PT_UDP                 17
#define PT_IPV6                41
#define PT_GRE                 47  
#define PT_PIM                 103
#define PT_MAC                 200
#define PT_MAC802_11           201
#define PT_RAW                 255

#define PT_SUB_ND              1
#define PT_SUB_REQ             2
#define PT_SUB_REP             3
#define PT_SUB_DAT             4
#define PT_SUB_ACK             5

#define PK_BROADCAST_ADDR      4294967295U
#define PK_TTL_DEFAULT         32

#define PK_CACHE_SIZE          48
//#define PK_CACHE_EXPIRED       10  // in ms

typedef unsigned int packet_id_t;
typedef unsigned char packet_type_t;
typedef unsigned int seqno_t;

using namespace std;

class Packet {
    DEBUG_INIT();

private:
    static packet_id_t tNextId;

private:
    packet_id_t tId;               // Unique packet id
    node_id_t tSrc;                // Source node id
    node_id_t tDst;                // Destination node id
    packet_type_t tType;           // Packet type
    packet_type_t tSubType;        // Packet sub type
    int iTTL;                      // TTL of the packet
    seqno_t tSeqNo;                // Sequence #
    time_t tTimeStamp;             // Creation time

public:
    Packet(node_id_t tSrc,
	   node_id_t tDst,
	   packet_type_t tType,
	   packet_type_t tSubType = 0,
	   int iTTL = PK_TTL_DEFAULT,
	   seqno_t tSeqNo = 0);
    Packet(const Packet &p);
    static packet_id_t nextId(void) {
	return (tNextId++);
    }
    virtual ~Packet(void) {
	DEBUG("%d destroyed", id());
    }
    packet_id_t id(void) const {
	return (tId);
    }
    node_id_t src(void) const {
	return (tSrc);
    }
    node_id_t dst(void) const {
	return (tDst);
    }
    packet_type_t type(void) const {
	return (tType);
    }
    packet_type_t subType(void) const {
	return (tSubType);
    }
    int ttl(void) const {
	return (iTTL);
    }
    void decTTL(void) {
	iTTL--;
    }
    seqno_t seqNo(void) const {
	return (tSeqNo);
    }
    time_t timeStamp(void) const {
	return (tTimeStamp);
    }
    // Destroy a packet
    void destroy(void) {
	delete this;
    }
    // Check if the packet is destined to the node
    bool forMe(Node *pcNode) {
	return (dst() == pcNode->id() || dst() == PK_BROADCAST_ADDR);
    }
    // Clone a packet which will have a new packet id
    virtual Packet *clone(void) {
	return (new Packet(*this));
    }
    // Compare the packet
    virtual bool operator==(const Packet &p) const {
	return (id() == p.id());
    }
};

class Packets {
protected:
    int iMaxCacheSize;

public:
    Packets(void) : iMaxCacheSize(0) {
    }
    int maxCacheSize(void) const {
	return (iMaxCacheSize);
    }
    virtual void clear(time_t tNow) = 0;
    virtual void cache(Packet *pcPacket) = 0;
    virtual bool hit(Packet *pcPacket) = 0;
};

// XXX: need better algorithm for packet caching
class PacketsFIFO : public Packets, public list<Packet *> {
public:
    void clear(time_t tNow) {
#ifdef PK_CACHE_EXPIRED
	for (iterator itr = begin(); itr != end(); ) {
	    if (tNow - (*itr)->timeStamp() > PK_CACHE_EXPIRED) {
		// Destroy expired cloned packets
		iterator itr2 = itr;
		itr++;
		(*itr2)->destroy();
		erase(itr2);
	    } else {
		itr++;
	    }
	}
#else
	for (iterator itr = begin(); itr != end(); ) {
	    (*itr)->destroy();
	    itr++;
	}
	list<Packet *>::clear();
#endif
    }
    void cache(Packet *pcPacket) {
	if (size() == PK_CACHE_SIZE) {
	    front()->destroy();
	    pop_front();
	}
	// Clone the packet before caching it
	push_back(pcPacket->clone());
	if (size() > iMaxCacheSize) {
	    iMaxCacheSize = size();
	}
    }
    bool hit(Packet *pcPacket) {
	for (iterator itr = begin(); itr != end(); itr++) {
	    if (**itr == *pcPacket) {
		return (true);
	    }
	}
	return (false);
    }
};

#endif // PACKET_H
