// $Id: torrent_node.cc,v 1.9 2005/06/13 06:06:53 weesan Exp $

#include <stdlib.h>
#include <assert.h>
#include "torrent_node.h"
#include "simulation.h"
#include "packet.h"

#define TN_SINK_NODE()        (id() == 0)

#define TN_REQ_BACKOFF_TIMER  SIM_1_SEC
// Make sure the 'retry' timer is at least >= TN_REQ_BACKOFF_TIMER
#define TN_RETRY_TIMER        TN_REQ_BACKOFF_TIMER

DEBUG_REGISTER(TorrentNode);

static void torrent_node_broadcast(void *pvNode, void *pvPacket, void *) {
    TorrentNode *pcNode = (TorrentNode *)pvNode;
    TorrentReq *pcTorrentReq = (TorrentReq *)pvPacket;

    // If REQ suppression is turned on and the REQ has been cached, do nothing
    if (pcNode->reqSuppression() && pcNode->fwdedPkts()->hit(pcTorrentReq)) {
	pcTorrentReq->destroy();
	return;
    }

    // Otherwise, go ahead and broadcast the REQ packet
    pcNode->broadcast(pcTorrentReq);
    pcNode->incNoofReq();
}

static void torrent_node_retry_timer(void *pvNode, void *, void *) {
    TorrentNode *pcNode = (TorrentNode *)pvNode;

    if (!pcNode->done()) {
	if (pcNode->page() <= pcNode->pageRetry() &&
	    pcNode->cell() <= pcNode->cellRetry()) {
	    // Request it again since REQ might be lost
	    AT(0, torrent_node_broadcast, pcNode,
	       new TorrentReq(pcNode->id(), pcNode->nodeType(),
			      pcNode->page(), pcNode->cell()));
	} else {
	    // Bring the retry machanism up-to-dated
	    pcNode->updateRetry();
	}
	// Keep monitoring the requests
	AT(TN_RETRY_TIMER, torrent_node_retry_timer, pcNode);
    }
}

///////////////////////////////////

TorrentNode::TorrentNode(int iX, int iY, int iNodeType,
			 bool bReqSuppression,
			 bool bLocalRecovery,
			 bool bPushMore,
			 Packets *pcFwdedPkts,
			 unsigned int uiBandwidth, int iR, Queue *pcQueue) :
    WirelessNode(iX, iY, uiBandwidth, iR, pcQueue, pcFwdedPkts),
    iNodeType(iNodeType),
    iPage(0),
    iCell(0),
    iPageRetry(0),
    iCellRetry(0),
    tDone(0),
    iNoofReq(0),
    bReqSuppression(bReqSuppression),
    bLocalRecovery(bLocalRecovery),
    bPushMore(bPushMore) {
    // Sink does not request for new image
    if (!TN_SINK_NODE()) {
	// In order to do aggregation and suppression, we randomly wait for a
	// few ms before asking
	time_t t = AT(JILTER(TN_REQ_BACKOFF_TIMER),
		      torrent_node_broadcast,
		      this,
		      new TorrentReq(id(), nodeType()));
	// Schedule a timer to keep track of the request above
	AT(t + TN_RETRY_TIMER, torrent_node_retry_timer, this);
    }
}

bool TorrentNode::expect(TorrentDat *pcTorrentDat) {
    return (pcTorrentDat->nodeType() == nodeType() &&
	    pcTorrentDat->page() == page() &&
	    pcTorrentDat->cell() == cell());
}

void TorrentNode::nextReq(TorrentDat *pcTorrentDat) {
    // Sink node do not request any packets, it just serves
    if (TN_SINK_NODE()) {
	return;
    }

    do {
	if (cell() + 1 < TN_CELL_MAX(page())) {
	    iCell++;
	} else if (page() + 1 < TN_PAGE_MAX) {
	    iPage++;
	    iCell = 0;
	} else {
	    DEBUG("%d received last DAT packet %d, "
		  "type %d page %d cell %d",
		  id(),
		  pcTorrentDat->id(), pcTorrentDat->nodeType(),
		  pcTorrentDat->page(), pcTorrentDat->cell());
	    // I am done now
	    done(NOW());
	    // Do not schedule any more REQ
	    break;
	}
	// Schedule one more REQ
	AT(JILTER(TN_REQ_BACKOFF_TIMER), torrent_node_broadcast, this,
	   new TorrentReq(id(), nodeType(), page(), cell()));
    } while (false);
}

bool TorrentNode::haveDat(TorrentReq *pcTorrentReq) {
    TorrentDat cTorrentDat(pcTorrentReq->dst(),
			   pcTorrentReq->src(),
			   pcTorrentReq->id(),
			   pcTorrentReq->nodeType(), 
			   pcTorrentReq->page(),
			   pcTorrentReq->cell());
    if (TN_SINK_NODE()) {
	// Sink has everything
	DEBUG("REQ type %d reaches sink from node %d",
	      pcTorrentReq->nodeType(), pcTorrentReq->src());
	return (true);
    } else if (localRecovery() && fwdedPkts()->hit(&cTorrentDat)) {
	DEBUG("REQ type %d reaches %d from node %d, local recovery",
	      pcTorrentReq->nodeType(), id (), pcTorrentReq->src());
	return (true);
    } else {
	return (false);
    }
}

void TorrentNode::pushMoreDat(TorrentReq *pcTorrentReq) {
    int iPage = pcTorrentReq->page();
    int iCell = pcTorrentReq->cell();

    if (iCell + 1 < TN_CELL_MAX(page())) {
	iCell++;
    } else if (iPage + 1 < TN_PAGE_MAX) {
	iPage++;
	iCell = 0;
    } else {
	// Nothing more to push
	return;
    }

    TorrentDat cTorrentDat(pcTorrentReq->dst(),
			   pcTorrentReq->src(),
			   pcTorrentReq->id(),
			   pcTorrentReq->nodeType(),
			   iPage,
			   iCell);
    if (TN_SINK_NODE() ||
	(localRecovery() && fwdedPkts()->hit(&cTorrentDat))) {
	DEBUG("%d push one more data (%d, %d) to node %d\n",
	      id(), iPage, iCell, pcTorrentReq->src());
	broadcast(new TorrentDat(cTorrentDat));
    }
}

void TorrentNode::packetSnooping(Packet *pcPacket, Link *pcIncoming) {
    if (pcPacket->type() == PT_MAC802_11 &&
	pcPacket->subType() == PT_SUB_DAT) {
	// Check if we are expecting the same packet
	TorrentDat *pcTorrentDat = (TorrentDat *)pcPacket;
	if (expect(pcTorrentDat)) {
	    DEBUG("%d overheard DAT packet %d from node %d, "
		  "type %d page %d cell %d",
		  id(), pcTorrentDat->id(), pcTorrentDat->src(),
		  pcTorrentDat->nodeType(), pcTorrentDat->page(),
		  pcTorrentDat->cell());
	    nextReq(pcTorrentDat);
	}
    }
}

void TorrentNode::rx(Packet *pcPacket, Link *pcIncoming) {
    /*
     * If the packet is not for us, ask parent class to do the job.
     * Before doing that, we check if we are expecting the same packet
     */
    if (!pcPacket->forMe(this)) {
	// Packet snooping
	packetSnooping(pcPacket, pcIncoming);

	/*
	 * Since the packet is not for us, we need to ask parent class
	 * to forward the packet
	 */
	WirelessNode::rx(pcPacket, pcIncoming);
	return;
    }

    // If the packet is for me
    switch (pcPacket->type()) {
    case PT_MAC802_11:
	switch (pcPacket->subType()) {
	case PT_SUB_REQ: {
	    // Check if we could serve some nodes with the packets we have

	    TorrentReq *pcTorrentReq = (TorrentReq *)pcPacket;

	    DEBUG("%d received REQ packet %d from %d, type %d, page %d",
		  id(), pcTorrentReq->id(), pcTorrentReq->src(),
		  pcTorrentReq->nodeType(), pcTorrentReq->page());

	    /*
	     * In addition to ND packets, REQ packets can be used to setup
	     * gradient as well
	     */
	    setupGradient(pcTorrentReq->src(), pcIncoming->srcNode());

	    // Reply with DAT packet if we have it
	    if  (haveDat(pcTorrentReq)) {
		// Reply with requested data packet
		TorrentDat *pcTorrentDat =
		    new TorrentDat(id(),
				   pcTorrentReq->src(),
				   pcTorrentReq->id(),
				   pcTorrentReq->nodeType(),
				   pcTorrentReq->page(),
				   pcTorrentReq->cell());
		assert(pcTorrentDat != NULL);
		// Send DAT
		broadcast(pcTorrentDat);
		// Push more DAT if available
		if (pushMore()) {
		    pushMoreDat(pcTorrentReq);
		}
		// Request packet needs to be destroyed after this
	    } else {
		// Otherwise, simply broadcast the REQ
		broadcast(pcTorrentReq, pcIncoming);
		return;
	    }
	}
	case PT_SUB_DAT: {
	    TorrentDat *pcTorrentDat = (TorrentDat *)pcPacket;

	    DEBUG("%d received DAT packet %d from node %d, "
		  "type %d page %d cell %d",
		  id(), pcTorrentDat->id(), pcTorrentDat->src(),
		  pcTorrentDat->nodeType(), pcTorrentDat->page(),
		  pcTorrentDat->cell());
	    if (expect(pcTorrentDat)) {
		nextReq(pcTorrentDat);
	    }
	    break;
	}
	default:
	    break;
	}
	break;
    default:
	break;
    }

    // We terminate the torrent packets here!
    pcPacket->destroy();
}
