// $Id: torrent_node.h,v 1.10 2005/06/13 06:06:53 weesan Exp $

#ifndef TORRENT_NODE_H
#define TORRENT_NODE_H

#include <list>
#include "wireless_node.h"
#include "debug.h"

#define TN_TYPE_MAX       3

#define TN_IMAGE_SIZE     2000      // Arbitrary image size
#define TN_PAGE_SIZE      1104     // Same as Deluge
#define TN_CELL_SIZE      23       // Same as Deluge which calls it 'packet'

#define TN_PAGE_MAX       ((TN_IMAGE_SIZE + TN_PAGE_SIZE - 1) / TN_PAGE_SIZE)
#define TN_CELL_MAX(p)                                                   \
    (((((p) == TN_PAGE_MAX - 1) ?                                          \
      TN_IMAGE_SIZE % TN_PAGE_SIZE : TN_PAGE_SIZE) + TN_CELL_SIZE - 1) / \
     TN_CELL_SIZE)

class TorrentPacket : public Packet {
private:
    int iNodeType;                 // Node/Image type
    int iPage;                     // Requesting page
    int iCell;                     // Requesting cell within the page above

public:
    TorrentPacket(node_id_t tSrc, node_id_t tDst, packet_type_t tSubType,
		  int iNodeType, int iPage = 0, int iCell = 0) :
	Packet(tSrc, tDst, PT_MAC802_11, tSubType),
	iNodeType(iNodeType),
	iPage(iPage),
	iCell(iCell) {
    }
    TorrentPacket(const TorrentPacket &t) :
	Packet(t),
	iNodeType(t.nodeType()),
	iPage(t.page()),
	iCell(t.cell()) {
    }
    int nodeType(void) const {
	return (iNodeType);
    }
    int page(void) const {
	return (iPage);
    }
    int cell(void) const {
	return (iCell);
    }
    virtual Packet *clone(void) = 0;
    // We don't care about the src and dst
    bool operator==(const Packet &p) const {
	const TorrentPacket &q = (const TorrentPacket &)p;
	return (type()     == q.type() &&
		subType()  == q.subType() &&
		seqNo()    == q.seqNo() &&
		nodeType() == q.nodeType() &&
		page()     == q.page() &&
		cell()     == q.cell());
    }
};

class TorrentReq : public TorrentPacket {
public:
    TorrentReq(node_id_t tSrc, int iNodeType, int iPage = 0, int iCell = 0) :
	TorrentPacket(tSrc, PK_BROADCAST_ADDR, PT_SUB_REQ,
		      iNodeType, iPage, iCell) {
    }
    Packet *clone(void) {
	return (new TorrentReq(*this));
    }
};

class TorrentDat : public TorrentPacket {
private:
    packet_id_t tOrigPktId;        // Corresponding to the original req packet

public:
    TorrentDat(node_id_t tSrc, node_id_t tDst, packet_id_t tOrigPktId,
	       int iNodeType, int iPage, int iCell) :
	TorrentPacket(tSrc, tDst, PT_SUB_DAT, iNodeType, iPage, iCell),
	tOrigPktId(tOrigPktId) {
    }
    TorrentDat(const TorrentDat &t) :
	TorrentPacket(t),
	tOrigPktId(t.origPktId()) {
    }
    packet_id_t origPktId(void) const {
	return (tOrigPktId);
    }
    Packet *clone(void) {
	return (new TorrentDat(*this));
    }
    bool operator==(const Packet &p) const {
	const TorrentDat &q = (const TorrentDat &)p;
	return (TorrentPacket::operator==(p) && origPktId() == q.origPktId());
    }
};

// Implement on-demand caching by using packet buckets for different type of
// packets
class TorrentCache : public Packets, vector<list<Packet *> > {
private:
    int iTotalPacket;

public:
    TorrentCache(void) :
	iTotalPacket(0) {
	for (int i = 0; i < TN_TYPE_MAX; i++) {
	    list<Packet *> l;
	    push_back(l);
	}
    }
    void clear(time_t tNow) {
#ifdef PK_CACHE_EXPIRED
	for (int i = 0; i < size(); i++) {
	    for (list<Packet *>::iterator itr = (*this)[i].begin();
		 itr != (*this)[i].end(); ) {
		if (tNow - (*itr)->timeStamp() > PK_CACHE_EXPIRED) {
		    // Destroy expired cloned packets
		    list<Packet *>::iterator itr2 = itr;
		    itr++;
		    (*itr2)->destroy();
		    (*this)[i].erase(itr2);
		    iTotalPacket--;
		} else {
		    itr++;
		}
	    }
	}
#else
	for (int i = 0; i < size(); i++) {
	    for (list<Packet *>::iterator itr = (*this)[i].begin();
		 itr != (*this)[i].end();
		 itr++) {
		// Destroy all cloned packets
		(*itr)->destroy();
	    }
	    (*this)[i].clear();
	}
	iTotalPacket = 0;
#endif
    }
    void cache(Packet *pcPacket) {
	TorrentPacket *pcTorrentPacket = (TorrentPacket *)pcPacket;
	int iType = pcTorrentPacket->nodeType();

	if (iType >= size()) {
	    return;
	}

	if (iTotalPacket == PK_CACHE_SIZE) {
	    /*
	     * Check the size of this type of packet cache, if it's empty,
	     * kick out the one next to it.
	     */
	    int iNewType = iType;
	    for (int i = 0; i < size(); i++) {
		list<Packet *> &k = (*this)[iNewType++ % size()];
		if (k.size()) {
		    k.front()->destroy();
		    k.pop_front();
		    iTotalPacket--;
		    break;
		}
	    }
	}

	list<Packet *> &l = (*this)[iType];

	// Clone the packet before caching it
	l.push_back(pcPacket->clone());
	if (++iTotalPacket > iMaxCacheSize) {
	    iMaxCacheSize = iTotalPacket;
	}
    }
    bool hit(Packet *pcPacket) {
	TorrentPacket *pcTorrentPacket = (TorrentPacket *)pcPacket;
	int iType = pcTorrentPacket->nodeType();

	if (iType >= size()) {
	    return (false);
	}

	for (list<Packet *>::iterator itr = (*this)[iType].begin();
	     itr != (*this)[iType].end();
	     itr++) {
	    if (**itr == *pcPacket) {
		return (true);
	    }
	}

	return (false);
    }
};

class TorrentNode : public WirelessNode {
    DEBUG_INIT();

private:
    int iNodeType;                 // Same as TorrentPacket
    int iPage;                     // Same as TorrentPacket
    int iCell;                     // Same as TorrentPacket
    int iPageRetry;                // Keep a copy in case we need to retry 
    int iCellRetry;                // Keep a copy in case we need to retry 
    time_t tDone;                  // The time when last DAT packet is received
    int iNoofReq;                  // The number of REQ actually sent
    bool bReqSuppression;          // REQ suppression or not?
    bool bLocalRecovery;           // Local recovery?
    bool bPushMore;                // Push more DAT if available?

private:
    bool expect(TorrentDat *pcTorrentDat);
    void nextReq(TorrentDat *pcTorrentDat);
    bool haveDat(TorrentReq *pcTorrentReq);
    void pushMoreDat(TorrentReq *pcTorrentReq);
    void packetSnooping(Packet *pcPacket, Link *pcIncoming);

public:
    TorrentNode(int iX, int iY, int iNodeType,
		bool bReqSuppression = true,
		bool bLocalRecovery = false,
		bool bPushMore = false,
		Packets *pcFwdedPkts = NULL,
		unsigned int uiBandwidth = LINK_19K,
		int iR = WL_RADIO_RANGE_DEFAULT,
		Queue *pcQueue = NULL);
    int nodeType(void) const {
	return (iNodeType);
    }
    int page(void) const {
	return (iPage);
    }
    int cell(void) const {
	return (iCell);
    }
    int pageRetry(void) const {
	return (iPageRetry);
    }
    int cellRetry(void) const {
	return (iCellRetry);
    }
    void updateRetry(void) {
	iPageRetry = iPage;
	iCellRetry = iCell;
    }
    time_t done(void) const {
	return (tDone);
    }
    void done(time_t tNow) {
	tDone = tNow;
    }
    int noofReq(void) const {
	return (iNoofReq);
    }
    void incNoofReq(void) {
	iNoofReq++;
    }
    bool reqSuppression(void) const {
	return (bReqSuppression);
    }
    bool localRecovery(void) const {
	return (bLocalRecovery);
    }
    bool pushMore(void) const {
	return (bPushMore);
    }
    virtual void rx(Packet *pcPacket, Link *pcIncoming);
};

#endif // TORRENT_NODE_H
