// $Id: wireless_node.cc,v 1.8 2005/06/12 09:31:10 weesan Exp $

#include <math.h>
#include <assert.h>
#include "wireless_node.h"
#include "wireless_link.h"
#include "simulation.h"
#include "packet.h"

// Comment out below would turn on periodic Neighbor Discovery mechanism
//#define WL_ND                     SIM_1_SEC
//#define WL_FLUSH_CACHE_TIMER      (SIM_1_SEC * 5)
#define WL_FLUSH_CACHE_TIMER      (SIM_1_MS * 10)

DEBUG_REGISTER(WirelessNode);

// Static functions

#ifdef WL_ND
static void wireless_node_ND(void *pvNode, void *pvPacket, void *) {
    WirelessNode *pcNode = (WirelessNode *)pvNode;
    Packet *pcPacket = (Packet *)pvPacket;

    pcNode->broadcast(pcPacket);

    /*
     * Here, we do not use JILTER macro because the code in WirelessNode
     * constructor should have already scattered ND message evenly
     */
    AT(WL_ND, wireless_node_ND, pcNode,
       new Packet(pcNode->id(), PK_BROADCAST_ADDR, PT_MAC802_11, PT_SUB_ND));
}
#endif

static void wireless_node_broadcast(void *pvNode,
				    void *pvPacket,
				    void *pvLink) {
    WirelessNode *pcNode = (WirelessNode *)pvNode;
    Packet *pcPacket = (Packet *)pvPacket;
    Link *pcLink = (Link *)pvLink;

    pcNode->broadcast(pcPacket, pcLink);
}

#ifdef WL_FLUSH_CACHE_TIMER
static void wireless_node_flush_cache(void *pvNode, void *, void *) {
    WirelessNode *pcNode = (WirelessNode *)pvNode;
    // Flush the cache
    pcNode->fwdedPkts()->clear(NOW());
    // Reschedule itself again
    AT(WL_FLUSH_CACHE_TIMER, wireless_node_flush_cache, pcNode);
}
#endif

///////////////////////////////////

WirelessNode::WirelessNode(int iX, int iY, unsigned int uiBandwidth,
			   int iR, Queue *pcQueue, Packets *_pcFwdedPkts) :
    iX(iX),
    iY(iY),
    iR(iR),
    uiBandwidth(uiBandwidth),
    pcFwdedPkts(_pcFwdedPkts),
    Node(pcQueue) {
    if (pcFwdedPkts == NULL) {
	pcFwdedPkts = new PacketsFIFO;
	assert(pcFwdedPkts != NULL);
    }

#ifdef WL_ND
    // Using JILTER macro to scatter the ND for each wireless node
    AT(JILTER(WL_ND), wireless_node_ND, this,
       new Packet(id(), PK_BROADCAST_ADDR, PT_MAC802_11, PT_SUB_ND));
#endif

#ifdef WL_FLUSH_CACHE_TIMER
    AT(JILTER(WL_FLUSH_CACHE_TIMER), wireless_node_flush_cache, this);
#endif
}

void WirelessNode::setupGradient(node_id_t tDst, Node *pcNeighborNode) {
    assert(pcNeighborNode != NULL);

    // Try to find if there is a link from itself to the neighbor node
    Link *pcOutgoing = cLinks.find(this, pcNeighborNode);
    if (!pcOutgoing) {
	pcOutgoing = new WirelessLink(this,  pcNeighborNode, bandwidth());
	assert(pcOutgoing != NULL);
    }
    
    // We can now get to the 'tDst' thru 'pcOutgoing'
    routingTable()->addEntry(tDst, pcOutgoing);
}

bool WirelessNode::inRange(WirelessNode *pcNode) {
    int dx = x() - pcNode->x();
    int dy = y() - pcNode->y();
    float fDistance = sqrtf(dx * dx + dy * dy);

    return ((int)fDistance <= r());
}

// NOTE: After being broadcasted, the original packet will be destroyed
void WirelessNode::broadcast(Packet *pcPacket, Link *pcIncoming) {
    DEBUG("%d broadcasting packet %d, ttl %d",
	  id(), pcPacket->id(), pcPacket->ttl());

    // We don't want to drop local originated packets with TTL of 1
    if (pcIncoming != NULL) {
	pcPacket->decTTL();
    }

    // Drop the packet if ttl <= 0
    if (pcPacket->ttl() <= 0) {
	pcPacket->destroy();
	return;
    }

    // Check if the packet is being broadcasted before
    if (fwdedPkts()->hit(pcPacket)) {
	DEBUG("%d detected dup packet %d, withdraw broadcast",
	      id (), pcPacket->id());
	return;
    } else {
	fwdedPkts()->cache(pcPacket);
    }

    for (Nodes::iterator itr = cNodes.begin(); itr != cNodes.end(); itr++) {
	// XXX: Assuming all nodes are either wired or wireless but not both
	WirelessNode *pcDstNode = (WirelessNode *)*itr;
	if (pcDstNode->id() != id() && inRange(pcDstNode)) {
	    /*
	     * Need to clone the packet since Packet is just a based class.
	     * Also, current packet will be dropped after being broadcasted
	     */
	    Packet *pcNewPacket = pcPacket->clone();
	    if (pcNewPacket) {
		/*
		 * Need to make sure there exists a wireless link between
		 * itself and the destination node before broadcasting
		 */
		setupGradient(pcDstNode->id(), pcDstNode);
		tx(pcNewPacket, pcDstNode, pcIncoming);
	    }
	}
    }

    // The packet is destroyed after done broadcasting it
    pcPacket->destroy();
}

// Receive a packet from a link
void WirelessNode::rx(Packet *pcPacket, Link *pcIncoming) {
    // If the packet is not for me, re-broadcast it
    if (!pcPacket->forMe(this)) {
#ifdef WL_BROADCAST_BACKOFF_TIMER
	AT(JILTER(WL_BROADCAST_BACKOFF_TIMER), wireless_node_broadcast, this,
	   pcPacket, pcIncoming);
#else
	broadcast(pcPacket, pcIncoming);
#endif
	return;
    }

    // If the packet is for me
    switch (pcPacket->type()) {
    case PT_MAC802_11:
	switch (pcPacket->subType()) {
	case PT_SUB_ND: {
	    DEBUG("%d received ND packet %d from %d",
		  id(), pcPacket->id(), pcPacket->src());
	    // Setup a gradient back to the origin of the ND packet
	    setupGradient(pcPacket->src(), pcIncoming->srcNode());
	    break;
	}
	default:
	    DEBUG("Invalid packet subtype: %d", pcPacket->subType());
	    break;
	}
	break;
    default:
	DEBUG("Invalid packet type: %d", pcPacket->type());
	break;
    }

    // We terminate the wireless packets here!
    pcPacket->destroy();
}
